-------------------------------------------------------------------
-- KSI Labs, LLC CENTIPEDE-PCI Mach 3 specific stuff.            --
--                                                               --
-- This is Mach 3 specific stuff.                                --
--                                                               --
-- This is the main part, the glue that connects all the bits    --
-- and pieces together and then it is included into the top      --
-- level file, centipede_pci, as a component.                    --
-------------------------------------------------------------------

--------------------------------------------------------------------------
--			LICENSE and Terms of Use.			--
--									--
-- First of all, this is NOT GPL or Free Software. It is open so you	--
-- have the VHDL source code. There is also the source code for the	--
-- Windows Engine for Mach3 CNC software available from KSI Labs, LLC	--
-- website, http://www.ksilabs.com.					--
--									--
-- You are free to use this particular VHDL code, make modifications to	--
-- customize it for your own needs, fix inevitable bugs, port it to 	--
-- other CNC software and do anything else as long as your derivative	--
-- work that is based on this code or uses parts or pieces thereof is	--
-- made to be used with any KSI Labs, LLC products.			--
--									--
-- We do NOT grant you a permission to use any parts of this code in	--
-- any products that are designed for products not manufactured or	--
-- distributed by KSI Labs, LLC. We might give such a permission on	--
-- case-by-case basis but it is not legal to use this code or bits and	--
-- pieces thereof with products not coming from KSI Labs, LLC without	--
-- our explicit written permission.					--
--									--
-- The original copyright notices must be included in all derivative	--
-- works. You do NOT have to ask our permission to make any changes to	--
-- this code and you are not under obligation to give us any of your	--
-- work back as long as you are using it for your own needs.		--
--									--
-- However we do NOT give our permission to distribute your derivative	--
-- work unless all original copyrights are retained, all the conditions	--
-- of this license are met (use only with KSI Labs, LLC products) and	--
-- full source code for such derivative work is available for free to	--
-- everybody without any further conditions including but not limited	--
-- to signing any kinds of NDA. We explicitely deny any permission to	--
-- distribute any derivative work based on this code with closed source	--
-- binary-only fragments or include bits or pieces of this code in such	--
-- products. That essentially means that it is NOT legal to use it with --
-- ANY proprietary HDL megafunctions, megacores, etc. if the resulting	--
-- product is distributed. It IS legal to use whatever you want if the	--
-- resulting product is made for use with a KSI Labs, LLC product AND	--
-- that resulting product is not distributed. Exclusions can be granted --
-- by KSI Labs, LLC on case-by-case basis.				--
--------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.mach3_data_types.all;
use work.mach3_bits.all;


entity mach3_main is
generic (
	fifo_depth	: natural := 20
);
port (
	pulser_bus	: out std_logic_vector(11 downto 0);
	sp_dir		: out std_logic;
	sp_step		: out std_logic;
	axes_active	: out std_logic_vector(5 downto 0);
	irq_home	: out std_logic_vector(5 downto 0);
	irq_probe	: out std_logic;
	irq_cycle	: out std_logic;
	irq_done	: out std_logic;
	irq_fault	: out std_logic;
	irq_fovf	: out std_logic;
	idx		: out index_regs;
	cfg_reg		: out std_logic_vector(31 downto 0);
	cfg2_reg	: out std_logic_vector(31 downto 0);
	cfg3_reg	: out std_logic_vector(7 downto 0);
	ctl_reg		: out std_logic_vector(15 downto 0);
	stat_reg	: out std_logic_vector(31 downto 0);
	enc_reg		: out std_logic_vector(31 downto 0);
	spcnt_reg	: out std_logic_vector(27 downto 0);
	stspctl_reg	: out std_logic_vector(26 downto 0);
	sp_idx_pulse	: out std_logic;
	sp_tim_pulse	: out std_logic;
	cpump_out	: out std_logic;
	fovf_clr	: in std_logic;
	enc_data	: in std_logic_vector(7 downto 0);
	data_in		: in std_logic_vector(31 downto 0);
	din_reg		: in std_logic_vector(31 downto 0);
	idx_sel		: in std_logic_vector(5 downto 0);
	cfg_sel		: in std_logic;
	cfg2_sel	: in std_logic;
	cfg3_sel	: in std_logic;
	ctl_sel		: in std_logic;
	stat_sel	: in std_logic;
	fifo_sel	: in std_logic;
	enc_sel		: in std_logic;
	spctl_sel	: in std_logic;
	wd_restart	: in std_logic;
	rd		: in std_logic;
	wr		: in std_logic;
	clk		: in std_logic;
	nrst		: in std_logic;
	m_rst		: out std_logic
);

	function make_bb_step(seq: in unsigned(1 downto 0)) return std_logic_vector is
	begin
		case seq is
			when "00" => return "00";
			when "01" => return "10";
			when "10" => return "11";
			when "11" => return "01";
		end case;
	end make_bb_step;

end mach3_main;


architecture guts of mach3_main is

	-------------
	-- Signals --
	-------------

	type bb_steps_type is array (0 to 5) of unsigned(1 downto 0);
	signal bb_steps		: bb_steps_type;

	type bb_st_type is array (0 to 5) of std_logic_vector(1 downto 0);
	signal bb_st		: bb_st_type;

	signal pulses_stepdir	: std_logic_vector(11 downto 0);
	signal pulses_bbwave	: std_logic_vector(11 downto 0);

	signal mach3_cfg_reg	: std_logic_vector(31 downto 0);
	signal mach3_cfg2_reg	: std_logic_vector(31 downto 0);
	signal mach3_cfg3_reg	: std_logic_vector(7 downto 0);
	signal mach3_ctl_reg	: std_logic_vector(15 downto 0);
	signal mach3_stat_reg	: std_logic_vector(31 downto 0);
	signal mach3_spcnt_reg	: std_logic_vector(27 downto 0);
	signal mach3_stsp_reg	: std_logic_vector(26 downto 0);

	signal idx_write	: std_logic_vector(5 downto 0);
	signal cfg_write	: std_logic;
	signal cfg2_write	: std_logic;
	signal cfg3_write	: std_logic;
	signal ctl_write	: std_logic;
	signal stat_write	: std_logic;
	signal stsp_write	: std_logic;

	signal enc_rd		: std_logic;

	signal reset		: std_logic;	-- Active high

	signal idx_regs		: index_regs;

	signal axis_en		: std_logic_vector(5 downto 0);

	signal lim_pos_in	: std_logic_vector(5 downto 0);
	signal lim_pos_hit	: std_logic_vector(5 downto 0);
	signal lim_pos_clr	: std_logic_vector(5 downto 0);
	signal lim_neg_in	: std_logic_vector(5 downto 0);
	signal lim_neg_hit	: std_logic_vector(5 downto 0);
	signal lim_neg_clr	: std_logic_vector(5 downto 0);
	signal home_en		: std_logic_vector(5 downto 0);
	signal home_sw_in	: std_logic_vector(5 downto 0);
	signal home_hit		: std_logic_vector(5 downto 0);
	signal home_clr		: std_logic_vector(5 downto 0);

	signal dirs_out		: std_logic_vector(5 downto 0);
	signal dirs_inv		: std_logic_vector(5 downto 0);
	signal steps_out	: std_logic_vector(5 downto 0);
	signal steps_inv	: std_logic_vector(5 downto 0);

	signal fifo_push	: std_logic;
	signal fifo_pop		: std_logic;
	signal fifo_purge	: std_logic;
	signal step_on		: std_logic;
	signal step_stb		: std_logic;
	signal fifo_out		: std_logic_vector(11 downto 0);
	signal fifo_in		: std_logic_vector(11 downto 0);
	signal probe_hit	: std_logic;
	signal probe_clr	: std_logic;

	signal no_limits	: std_logic;

	signal simulate		: std_logic;

	signal probe_in		: std_logic;
	signal rst_in		: std_logic;

	signal estop_switch	: std_logic;
	signal estop_sw_in	: std_logic;
	signal estop_req_in	: std_logic;
	signal estop_lim_in	: std_logic;
	signal estop_clr	: std_logic;

	signal f_pop		: std_logic;
	
	signal stat_estop	: std_logic;
	signal stat_fault	: std_logic;

	signal f_room		: std_logic_vector(4 downto 0);
	signal f_full		: std_logic;
	signal f_empty		: std_logic;
	signal fsm_busy		: std_logic;
	signal fsm_busy_prev	: std_logic;

	signal req_cycle	: std_logic;
	signal req_done		: std_logic;
	signal req_fovf		: std_logic;

	signal f_ovf_clr	: std_logic;

	signal run_prev		: std_logic;
	signal run_cur		: std_logic;
	signal mach3_run	: std_logic;

	signal din_sp_idx	: std_logic;
	signal din_sp_tim	: std_logic;
	signal sp_idx_ena	: std_logic;
	signal sp_idx_rev	: std_logic;
	signal sp_tim_ena	: std_logic;
	signal sp_tim_rev	: std_logic;
	signal stsp_ena		: std_logic;
	signal sp_idx_deadtime	: std_logic_vector(9 downto 0);
	signal sp_tim_deadtime	: std_logic_vector(8 downto 0);

	signal s_sp_idx_pulse	: std_logic;
	signal s_sp_tim_pulse	: std_logic;
	signal s_sp_stopped	: std_logic;
	signal wait_for_spindle	: std_logic;
	
	signal cpump_clk	: std_logic;
	signal cpump_cntr	: unsigned(5 downto 0);
	signal cpump_cntr_eoc	: unsigned(5 downto 0);
	signal cpump_cntr_rst	: std_logic;
	signal cpump_out_sig	: std_logic;
	signal cpump_out_gated	: std_logic;
	
	signal wd_cntr		: unsigned(2 downto 0);
	signal wd_enable	: std_logic;
	signal wd_full		: std_logic;

begin

	--=====================--
	-- COMPONENT INSTANCES --
	--=====================--

	----------------------------
	-- Motion (Step/Dir) FIFO --
	----------------------------

	z1: component mach3_fifo
	generic map (
		fifo_depth => fifo_depth
	)
	port map (
		fifo_in => fifo_in,
		fifo_out => fifo_out,
		fifo_push => fifo_push,
		fifo_pop => fifo_pop,
		fifo_purge => fifo_purge,
		fifo_run => mach3_run,
		fifo_full => f_full,
		fifo_empty => f_empty,
		fifo_ovf_irq => req_fovf,
		fifo_ovf_clr => f_ovf_clr,
		fifo_room => f_room,
		clk => clk,
		rst => reset
	);

	---------------------------------------------
	-- Position (Index) registers and latches. --
	---------------------------------------------

	z2: component mach3_index
	port map (
		data_in => data_in,
		steps => steps_out,
		step_stb => step_stb,
		dirs => dirs_out,
		idx_regs => idx_regs,
		home_sw_hit => home_hit,
		probe_hit => probe_hit,
		idx_wr => idx_write,
		clk => clk,
		rst => reset
	);

	--------------------------------------------
	-- Home/Probe/Limit switches for all axes --
	--------------------------------------------

	z3: component mach3_switches
	port map (
		sw_probe => probe_in,
		sw_home => home_sw_in,
		sw_lim_neg => lim_neg_in,
		sw_lim_pos => lim_pos_in,
		lim_neg_hit => lim_neg_hit,
		lim_pos_hit => lim_pos_hit,
		probe_hit => probe_hit,
		home_hit => home_hit,
		lim_neg_clr => lim_neg_clr,
		lim_pos_clr => lim_pos_clr,
		probe_clr => probe_clr,
		home_clr => home_clr,
		clk => clk,
		rst => reset
	);

	---------------------
	-- Sequencer (FSM) --
	---------------------

	z4: component mach3_fsm
	port map (
		cycle_div => mach3_ctl_reg(15 downto 5),
		fifo_pop => f_pop,
		fifo_empty => f_empty,
		tmr_req => req_cycle,
		fsm_busy => fsm_busy,
		step_on => step_on,
		step_stb => step_stb,
		fsm_run => mach3_run,
		spindle_pulse => s_sp_idx_pulse,
		wait4spindle => wait_for_spindle,
		ctl_wr => ctl_write,
		clk => clk,
		rst => reset
	);

	--------------
	-- Encoders --
	--------------

	z5: component mach3_encoders
	port map (
		enc_rd => enc_rd,
		enc_data => enc_data,
		enc_counts => enc_reg,
		clk => clk,
		rst => reset
	);

	-------------
	-- Spindle --
	-------------

	z6: component mach3_spindle
	port map (
		sp_din_idx => din_sp_idx,
		sp_din_tim => din_sp_tim,
		sp_idx_ena => sp_idx_ena,
		sp_tim_ena => sp_tim_ena,
		stsp_ena => stsp_ena,
		idx_deadtime => sp_idx_deadtime,
		tim_deadtime => sp_tim_deadtime,
		sp_cnt => mach3_spcnt_reg,
		stsp_ctl => mach3_stsp_reg,
		sp_idx_pulse => s_sp_idx_pulse,
		sp_tim_pulse => s_sp_tim_pulse,
		sp_stopped => s_sp_stopped,
		dir_out => sp_dir,
		step_out => sp_step,
		clk520khz => cpump_clk,
		stsp_wr => stsp_write,
		data_in => data_in,
		clk => clk,
		rst => reset
	);

	-----------------------------
	-- Hardware/software reset --
	-----------------------------
	reset_proc : process (nrst, clk)
	begin
		if (nrst='0') then
			reset <= '1';
		elsif (rising_edge(clk)) then
			if (rst_in='1') then
				reset <= '1';
			else
				reset <= '0';
			end if;
		end if;
	end process;


	----------------------------------
	-- Configuration register write --
	----------------------------------
	cfg_write_proc : process (reset, clk)
	begin
		if (reset='1') then
			mach3_cfg_reg <= (others => '0');
		elsif (rising_edge(clk)) then
			if (cfg_write='1') then
				mach3_cfg_reg <= data_in;
			end if;
		end if;
	end process;


	-----------------------------------------
	-- Second configuration register write --
	-----------------------------------------
	cfg2_write_proc : process (reset, clk)
	begin
		if (reset='1') then
			mach3_cfg2_reg <= (others => '0');
		elsif (rising_edge(clk)) then
			if (cfg2_write='1') then
				mach3_cfg2_reg <= data_in;
			end if;
		end if;
	end process;


	----------------------------------------
	-- Third configuration register write --
	----------------------------------------
	cfg3_write_proc : process (reset, clk)
	begin
		if (reset='1') then
			mach3_cfg3_reg <= (others =>'0');
		elsif (rising_edge(clk)) then
			if (cfg3_write='1') then
				mach3_cfg3_reg <= data_in(7 downto 0);
			end if;
		end if;
	end process;

	----------------------------
	-- Control register write --
	----------------------------
	ctl_write_proc : process (reset, clk)
	begin
		if (reset='1') then
			mach3_ctl_reg <= (others => '0');
		elsif (rising_edge(clk)) then
			if (ctl_write='1') then
				mach3_ctl_reg <= data_in(mach3_ctl_reg'range);
			end if;
		end if;
	end process;


	------------------------
	-- FSM Done Interrupt --
	------------------------
	req_delay : process (clk)
	begin
		if (rising_edge(clk)) then
			fsm_busy_prev <= fsm_busy;
		end if;
	end process;


	------------------
	-- EStop Switch --
	------------------
	estop_sw_proc : process (reset, clk)
	begin
		if (reset='1') then
			estop_sw_in <= '0';
		elsif (rising_edge(clk)) then
			if (estop_switch='0') then
				estop_sw_in <= '1';
			elsif (estop_clr='1') then
				estop_sw_in <= '0';
			end if;
		end if;
	end process;


	---------------------------
	-- Main run signal delay --
	---------------------------
	run_delay_proc : process(clk, reset)
	begin
		if (reset='1') then
			run_prev <= '0';
		elsif (rising_edge(clk)) then
			run_prev <= run_cur;
		end if;
	end process;


	---------------------------------------------------
	-- Bridgeport Series II Boss 6 Mill Step Signals --
	---------------------------------------------------
	bb_steps_proc : process(clk, reset)
	begin
		if (reset='1') then
			for i in 0 to 5 loop
				bb_steps(i) <= "00";
			end loop;
		elsif (rising_edge(clk)) then
			for i in 0 to 5 loop
				if (steps_out(i)='1' and step_stb='1') then
					if (dirs_out(i)='0') then
						bb_steps(i) <= bb_steps(i) + 1;
					else
						bb_steps(i) <= bb_steps(i) - 1;
					end if;
				end if;
			end loop;
		end if;
	end process;

	
	--------------------------------------------
	-- Charge Pump signals (~12kHz and ~5kHz) --
	--------------------------------------------
	cpump_train_proc : process(clk, reset)
	begin
		if (reset='1') then
			cpump_cntr <= (others => '0');
		elsif (rising_edge(clk)) then
			if (cpump_clk='1') then
				if (cpump_cntr_rst='1') then
					cpump_cntr <= (others => '0');
					cpump_out_sig <= not cpump_out_sig;
				else
					cpump_cntr <= cpump_cntr + 1;
				end if;
			end if;
		end if;
	end process;

	
	--------------------
	-- Watchdog timer --
	--------------------
	wd_timer_proc : process(clk, reset, wd_enable)
	begin
		if (reset='1' or wd_enable='0') then
			wd_cntr <= (others => '0');
		elsif (rising_edge(clk)) then
			if (wd_full='0') then
				if (wd_restart='1') then
					wd_cntr <= (others => '0');
				elsif (req_cycle='1') then
					wd_cntr <= wd_cntr + 1;
				end if;
			end if;
		end if;
	end process;

	
	----------------------------------
	-- Concurrent code and mappings --
	----------------------------------

	fifo_push <= fifo_sel and wr;
	fifo_purge <= ctl_write and data_in(30);
	fifo_in <= data_in(11 downto 0);
	fifo_pop <= f_pop and not stat_estop;

	idx_write_map : for i in 0 to 5 generate
		idx_write(i) <= idx_sel(i) and wr;
	end generate;

	cfg_write <= cfg_sel and wr;
	cfg2_write <= cfg2_sel and wr;
	cfg3_write <= cfg3_sel and wr;
	ctl_write <= ctl_sel and wr;
	stat_write <= stat_sel and wr;
	stsp_write <= spctl_sel and wr;

	enc_rd <= enc_sel and rd;
	
	axis_mapping : for i in 0 to 5 generate
		axis_en(i) <= mach3_cfg_reg(i*5);
		dirs_inv(i) <= mach3_cfg_reg(i*5+1);
		steps_inv(i) <= mach3_cfg2_reg(i);
		home_en(i) <= mach3_cfg_reg(i*5+4);

		lim_pos_clr(i) <= data_in(i*3) and stat_write;
		lim_neg_clr(i) <= data_in(i*3+1) and stat_write;
		home_clr(i) <= data_in(i*3+2) and stat_write;

		mach3_stat_reg(i*3) <= lim_pos_hit(i);
		mach3_stat_reg(i*3+1) <= lim_neg_hit(i);
		mach3_stat_reg(i*3+2) <= home_hit(i);

		lim_pos_in(i) <= (din_reg(i*3) xnor mach3_cfg_reg(i*5+2))
			and axis_en(i);
		lim_neg_in(i) <= (din_reg(i*3+1) xnor mach3_cfg_reg(i*5+2))
			and axis_en(i);
		home_sw_in(i) <= (din_reg(i*3+2) xnor mach3_cfg_reg(i*5+3))
			and axis_en(i) and home_en(i);

		dirs_out(i) <= fifo_out(i*2);
		steps_out(i) <= fifo_out(i*2+1);
--		steps_out(i) <= fifo_out(i*2+1) and axis_en(i);
	end generate;

	probe_in <= (din_reg(18) xnor mach3_cfg_reg(30)) and mach3_cfg_reg(31);
	probe_clr <= data_in(18) and stat_write;

	rst_in <= mach3_ctl_reg(4);

	mach3_stat_reg(18) <= probe_hit;
	mach3_stat_reg(23 downto 19) <= f_room;
	mach3_stat_reg(24) <= f_empty;
	mach3_stat_reg(25) <= f_full;

	mach3_stat_reg(26) <= '0';
	mach3_stat_reg(27) <= wd_full;

	mach3_stat_reg(28) <= s_sp_stopped;

	mach3_stat_reg(29) <= fsm_busy;

	no_limits <= mach3_ctl_reg(2);
	simulate <= mach3_ctl_reg(0);

	estop_switch <= din_reg(19);

	estop_req_in <= mach3_ctl_reg(1);
	estop_lim_in <= (lim_pos_hit(0) or lim_neg_hit(0) or
		lim_pos_hit(1) or lim_neg_hit(1) or
		lim_pos_hit(2) or lim_neg_hit(2) or
		lim_pos_hit(3) or lim_neg_hit(3) or
		lim_pos_hit(4) or lim_neg_hit(4) or
		lim_pos_hit(5) or lim_neg_hit(5)) and not no_limits;
	estop_clr <= data_in(30) and stat_write;

	pulser_stepdir : for i in 0 to 5 generate
		pulses_stepdir(i*2) <= dirs_out(i) xor dirs_inv(i);
		pulses_stepdir(i*2+1) <= ((steps_out(i) and step_on) xor steps_inv(i))
			when (stat_estop='0' and simulate='0') else steps_inv(i);
	end generate;

	pulser_bbwave : for i in 0 to 5 generate
		bb_st(i) <= make_bb_step(bb_steps(i));
		pulses_bbwave(i*2) <= bb_st(i)(0) when (dirs_inv(i)='0') else bb_st(i)(1);
		pulses_bbwave(i*2+1) <= bb_st(i)(1) when (dirs_inv(i)='0') else bb_st(i)(0);
	end generate;

	pulser_bus <= pulses_stepdir when (mach3_cfg3_reg(0)='0') else pulses_bbwave;
	
	stat_fault <= estop_sw_in or estop_lim_in or wd_full;
	stat_estop <= stat_fault or estop_req_in;

	mach3_stat_reg(30) <= stat_estop;

	cfg_reg <= mach3_cfg_reg;
	cfg2_reg <= mach3_cfg2_reg;
	cfg3_reg <= mach3_cfg3_reg;
	ctl_reg <= mach3_ctl_reg;
	stat_reg <= mach3_stat_reg;

	axes_active <= axis_en;
	idx <= idx_regs;

	f_ovf_clr <= fovf_clr;

	run_cur <= mach3_ctl_reg(3);
	mach3_run <= run_prev and run_cur;
	mach3_stat_reg(31) <= mach3_run;

	-- Spindle
	sp_idx_rev <= mach3_cfg2_reg(7);
	sp_tim_rev <= mach3_cfg2_reg(31);
	din_sp_idx <= din_reg(23) xor sp_idx_rev;
	din_sp_tim <= din_reg(22) xor sp_tim_rev;
	sp_idx_ena <= mach3_cfg2_reg(6);
	sp_tim_ena <= mach3_cfg2_reg(30);
	sp_idx_deadtime <= mach3_cfg2_reg(20 downto 11);
	sp_tim_deadtime <= mach3_cfg2_reg(29 downto 21);
	stsp_ena <= mach3_cfg2_reg(8);
	spcnt_reg <= mach3_spcnt_reg;
	stspctl_reg <= mach3_stsp_reg;
	sp_idx_pulse <= s_sp_idx_pulse;
	sp_tim_pulse <= s_sp_tim_pulse;
	wait_for_spindle <= data_in(31) and ctl_write;

	m_rst <= reset;

	-- Charge Pump
	cpump_cntr_eoc <= b"11_0111" when (mach3_cfg3_reg(3)='1') else b"01_0100";
	cpump_cntr_rst <= '1' when (cpump_cntr=cpump_cntr_eoc) else '0';
	cpump_out_gated <= cpump_out_sig when (mach3_run='1' and (mach3_cfg3_reg(2)='1' or stat_estop='0')) else '0';
	cpump_out <= cpump_out_gated;

	-- Watchdog
	wd_enable <= mach3_cfg3_reg(5);
	wd_full <= wd_cntr(2);
	
	----------
	-- IRQs --
	----------
	req_done <= fsm_busy_prev and not fsm_busy;

	irq_home <= home_hit;
	irq_probe <= probe_hit;
	irq_cycle <= req_cycle;
	irq_done <= req_done;
	irq_fovf <= req_fovf;

	irq_fault <= stat_fault;
end guts;
